# CornStarch
# by Eresse <eresse@eresse.net>

# CornStarch Module
module CornStarch

	# Application Controller
	class CornStarchController < ActionController::Base

		# CSRF Protection
		protect_from_forgery with: :exception

		# Global Filters
		before_filter :pagination_filter, :infiniscroll_filter, :locale_filter

		# Helper Methods
		helper_method :user_model, :current_user, :is_authed, :is_self, :paginate, :notice_msg

		# Use the Application's Layout
		layout 'layouts/application'

		# Include all Helpers
		helper CornStarch.helpers

		# Default URL Options
		def default_url_options options = {}

			# Include Locale in all generated URLs
			options.merge locale: I18n.locale
		end

		# Private
		private

		##################
		# Filter Methods #
		##################

		# Authentication Filter
		def auth_filter
			redirect_to corn_starch.new_session_url, notify_acs_denied('auth.login_required') unless is_authed
		end

		# Un-Authentication Filter
		def unauth_filter quiet = true
			redirect_to root_url, (quiet ? {} : notify_acs_denied('auth.already_logged_in')) if is_authed
		end

		# Self Filter
		def self_filter
			redirect_to root_url, notify_acs_denied('auth.invalid_user_id') unless is_self
		end

		# Not Self Filter
		def not_self_filter
			redirect_to root_url, notify_acs_denied('auth.invalid_user_id') if is_self
		end

		# Pagination Filter
		def pagination_filter

			# Create Pagination
			@pagination ||= {}
			session[:pagination] ||= {}

			# Set Records Per Page from Request Arguments
			session[:pagination][:recs_per_page] = params[:records_per_page].to_i if params[:records_per_page]

			# Init Pagination Vars
			@pagination[:current] = 0
			@pagination[:links] = []
			@pagination[:pg_count] = 0

			# Acquire Records per Page from Session / Configuration
			@pagination[:recs_per_page] = (session[:pagination][:recs_per_page] || CORNSTARCH_CONF[:pagination][:records_per_page][:default]).to_i
		end

		# InfiniScroll Filter
		def infiniscroll_filter

			# Create InfiniScroll
			@infiniscroll ||= {}
			@infiniscroll[:params] ||= {}

			# Set Search
			@infiniscroll[:params][:search] = params[:search] if params[:search]

			# Set Records Per Page
			@infiniscroll[:params][:records_per_page] = params[:records_per_page] if params[:records_per_page]
		end

		# Locale Filter
		def locale_filter

			# Set Locale
			I18n.locale = params[:locale] || I18n.default_locale
		end

		##################
		# Helper Methods #
		##################

		# User Model
		def user_model
			Kernel.const_get CORNSTARCH_CONF[:authentication][:user_model][:name]
		end

		# UID Field
		def uid_field
			CORNSTARCH_CONF[:authentication][:user_model][:uid_field]
		end

		# Current User
		def current_user
			@current_user ||= user_model.find(session[:user_id]) if session[:user_id]
		end

		# Is Authed
		def is_authed
			current_user != nil
		end

		# Is Self
		def is_self
			return false unless current_user
			params[:id] == current_user.id.to_s
		end

		# Get Records
		def get_records model, params, order = { id: :desc }
			paginate unisearch(model, params[:search]).order(order), params
		end

		# Render InfiniScroll Data
		def infiniscroll_data template, params = {}

			# Compute Next
			@infiniscroll[:next] = (@pagination[:current] < (@pagination[:pg_count] - 1)) ? @pagination[:current] + 1 : :none

			# Produce JSON
			params[:layout] = false
			{ json: { next: @infiniscroll[:next], content: render_to_string(template, params) } }
		end

		# Universal Search
		def unisearch model, search

			# Check Search
			return model unless search

			# Initialize Variables
			search_model = model
			search_q = ''
			search_a = []
			search_fields = {}
			first = true

			# Run through Model's Associations
			model.searchable_associations.each do |a|

				# Join into Search Model
				search_model = search_model.includes a

				# Acquire Association Target Model
				assoc = model.reflect_on_association(a).klass

				# Add Target Model's Search Fields
				search_fields[assoc.table_name] = assoc.searchable_attributes
			end

			# Add Base Model's Search Fields
			search_fields[model.table_name] = []
			model.searchable_attributes.each { |a| search_fields[model.table_name] << a }

			# Run through Search Fields Models
			search_fields.each do |m, fields|

				# Run through Model's Fields
				fields.each do |f|

					# Build Search Query & Arguments Array
					search_q = search_q + (first ? '' : 'OR ') + "#{m}.#{f} LIKE ? "
					search_a << "%#{search}%"
					first = false
				end
			end

			# Perform Search and eliminate doubles
			search_model.where(search_q, *search_a).references(model.table_name).uniq
		end

		# Paginate
		def paginate collection, params

			# Acquire Current Page
			current = (params[:current_page] || 0).to_i

			# Compute Page Count
			total = collection.count
			@pagination[:pg_count] = (total / @pagination[:recs_per_page]) + (((total % @pagination[:recs_per_page]) > 0) ? 1 : 0)

			# Compute Page Range
			range_start = current - (CORNSTARCH_CONF[:pagination][:page_range] / 2)
			range_start = 0 if range_start < 0
			range_end = current + (CORNSTARCH_CONF[:pagination][:page_range] / 2)
			range_end = @pagination[:pg_count] - 1 if range_end >= @pagination[:pg_count]

			# Create Page Links
			@pagination[:links] = range_start .. range_end
			@pagination[:current] = current

			# Fetch Records
			collection.limit(@pagination[:recs_per_page]).offset current * @pagination[:recs_per_page]
		end

		# Redirect to Result
		def redirect_to_result model, entity, action, res

			# Compute Result Status and Key
			key, status = res ? [:notice, :pass] : [:alert, :fail]

			# Acquire Singular Model Name
			model_name = model.model_name.singular

			# Generate URL Method
			url_meth = { create: ["new_#{model_name}_url".to_sym], update: ["edit_#{model_name}_url".to_sym, entity] }[action]

			# Generate URL
			url = send *url_meth

			# Redirect
			redirect_to (res ? entity : url), key => notice_msg(model, entity, action, status)
		end

		# Generic Notice Messages
		def notice_msg model, entity, action, result_or_name

			# Prepare Names
			uname = current_user ? current_user.display_name : ''
			ename = entity ? entity.display_name : ''

			# Acquire Errors
			errors = notify_errors entity

			# Acquire Translation for Model (if available)
			model = I18n.t("#{model.model_name.singular}.model_name".to_sym, default: model) if model

			# Switch on Action
			case action

				# Session Control
				when :login then notify_res 'gen_msg.login', result_or_name, name: html_escape(uname)
				when :logout then notify 'gen_msg.logout'

				# Create / Update Resources
				when :create then notify_res 'gen_msg.create', result_or_name, { model: model, name: html_escape(ename), errors: errors }, { model: model }
				when :update then notify_res 'gen_msg.update', result_or_name, { model: model, name: html_escape(ename), errors: errors }, { model: model }

				# Destroy Resources
				when :destroy then notify 'gen_msg.destroy', { model: model, name: html_escape(result_or_name) }, { model: model }

				# Unknown Action
				else notify 'gen_msg.unknown_action', action: html_escape(action)
			end
		end

		# Build Notification Error String
		def notify_errors entity
			entity.try(:errors).try(:full_messages).try(:collect) { |e| "\n - #{e}" }.try :join
		end

		# Build Notification
		def notify key, body_hash = nil, head_hash = nil
			{ head: t("#{key}.head", head_hash), body: t("#{key}.body", body_hash) }
		end

		# Build Pass / Fail Notification
		def notify_res key, res, body_hash = nil, head_hash = nil
			Hash[[:pass, :fail].map { |r| [r, notify("#{key}.#{r}", body_hash, head_hash)] }][res]
		end

		# Build Access-Denied Notification
		def notify_acs_denied key
			{ alert: { head: t('auth.access_denied'), body: t(key) } }
		end
	end
end
